#include "stdafx.h"

#include "custom_mem_mngr.h"

mm_context_t* mm_init_mem_mngr(int elements_num, size_t element_size, char* name)
{
    int i;
    mm_context_t *mem_mngr;

    mem_mngr = stf_malloc(mm_context_t, 1);
    if (!mem_mngr)
        return NULL;

    mem_mngr->name = name;

    mem_mngr->storage_ptr = stf_calloc(char, element_size * elements_num);
    if ( !(mem_mngr->storage_ptr) )
        return NULL;

    mem_mngr->free_pointers_storage = stf_malloc(void*, elements_num);
    if ( !(mem_mngr->free_pointers_storage) )
        return (stf_free(mem_mngr->storage_ptr), NULL);

    mem_mngr->free_pointer_ptr =  mem_mngr->free_pointers_storage + (elements_num-1);
    mem_mngr->end_point = mem_mngr->free_pointer_ptr;
    mem_mngr->element_size = element_size;
    mem_mngr->elements_number = elements_num;
    mem_mngr->max_elements_used = 0;
    mem_mngr->now_used_elements = 0;

    for (i = 0; i < elements_num; i++){
        mem_mngr->free_pointers_storage[i] = (void*)((char*)(mem_mngr->storage_ptr) + i * element_size);
    }
    return mem_mngr;
}

void mm_reset_mem_mngr(mm_context_t *mem_mngr)
{
    int elements_num = mem_mngr->elements_number;
    int element_size = mem_mngr->element_size;

    mem_mngr->free_pointer_ptr =  mem_mngr->free_pointers_storage + (elements_num-1);
    mem_mngr->end_point = mem_mngr->free_pointer_ptr;
    mem_mngr->max_elements_used = 0;
    mem_mngr->now_used_elements = 0;

    for (int i = 0; i < elements_num; i++){
        mem_mngr->free_pointers_storage[i] = (void*)((char*)(mem_mngr->storage_ptr) + i * element_size);
    }
}

void mm_free_mem_mngr(mm_context_t *mem_mngr)
{
    stf_free(mem_mngr->storage_ptr);
    stf_free(mem_mngr->free_pointers_storage);

    mem_mngr->free_pointer_ptr      = NULL;
    mem_mngr->free_pointers_storage = NULL;
    mem_mngr->storage_ptr           = NULL;
    mem_mngr->end_point             = NULL;
}

void *mm_get_mem(mm_context_t *mem_mngr)
{
#ifdef MEM_MNGR_SAFE_MODE
    if (mem_mngr->free_pointer_ptr < mem_mngr->free_pointers_storage){
        print("%s", mem_mngr->name);
        ASSERT("Memory full error !", 1);
    }
    mem_mngr->now_used_elements++;
    if (mem_mngr->now_used_elements > mem_mngr->max_elements_used)
        mem_mngr->max_elements_used = mem_mngr->now_used_elements;
#endif //MEM_MNGR_SAFE_MODE
    return *(mem_mngr->free_pointer_ptr--);
}

void mm_free_mem(mm_context_t *mem_mngr, void *mem)
{
#ifdef MEM_MNGR_SAFE_MODE
    if (mem_mngr->free_pointer_ptr >= mem_mngr->end_point){
        print("%s", mem_mngr->name);
        ASSERT("Error deallocating memory !", 1);
    }
    mem_mngr->now_used_elements--;
#endif //MEM_MNGR_SAFE_MODE

    *(++(mem_mngr->free_pointer_ptr)) = mem;
}

int mm_get_free_units_num(mm_context_t *mem_mngr)
{
    return (mem_mngr->free_pointer_ptr - mem_mngr->free_pointer_ptr + 1);
}

int mm_get_used_units_num(mm_context_t *mem_mngr)
{
    return (mem_mngr->end_point - mem_mngr->free_pointer_ptr);
}

mm_malloc_t *mm_init_malloc_manager(int part_size)
{
    mm_malloc_t *buff_mngr;
    buff_mngr = stf_malloc(mm_malloc_t, 1);
    buff_mngr->buff_count = 0;
    buff_mngr->buffers[0] = stf_malloc(char, part_size);
    buff_mngr->free_ptr = 0;
    buff_mngr->part_size = part_size;
    return buff_mngr;
}

void *mm_malloc_func(mm_malloc_t *mngr, int size)
{
    register void *ptr;
    if (mngr->free_ptr + size > mngr->part_size){
        ASSERT_IF((mngr->buff_count >= MM_BUFFERS_COUNT), "Not enough buffers !", 1);
        if ((mngr->part_size < size) || (mngr->part_size - mngr->free_ptr > mngr->part_size - size)){
            for(int i = mngr->buff_count; i >= 0; i--){
                mngr->buffers[i+1] = mngr->buffers[i];
            }
            mngr->buff_count++;
            mngr->buffers[0] = stf_malloc(char, size);
            return mngr->buffers[0];
        }
        mngr->buff_count++;
        mngr->buffers[mngr->buff_count] = stf_malloc(char, mngr->part_size);
        mngr->free_ptr = 0;
    }

    ptr = mngr->buffers[mngr->buff_count] + mngr->free_ptr;
    mngr->free_ptr += size + 3;
    mngr->free_ptr &= (~3);
    return ptr;
}

void *mm_calloc_func(mm_malloc_t *mngr, int size)
{
    register int *ptr = (int*)mm_malloc_func(mngr, size);
    register int *tmp = ptr;
    register int *end_ptr;

    size += 3;
    size &= (~3);
    size /= 4;
    end_ptr = ptr + size;
    while ((*ptr++, ptr) < end_ptr) *ptr++ = 0;
    return tmp;
}

void mm_free_malloc_mngr(mm_malloc_t *mngr)
{
    for (int i = 0; i <= mngr->buff_count; stf_free(mngr->buffers[i++]));
}